/** creates `size` number of divs inside `container` */
function createGrid(container, width = 10, height = 20) {
    container.width = width
    container.height = height
    container.cells = []

    container.setCell = (pos, c, remove = false) => {
        let i = pos.y * container.width + pos.x
        if (container.cells[i] === undefined) return
        remove ?
            container.cells[i].classList.remove(c) :
            container.cells[i].classList.add(c)
    }

    container.getCell = pos => {
        return container.cells[pos.y * container.width + pos.x]
    }

    container.markEdges = () => {
        for (let y = 0; y < container.height; y++) {
            for (let x = 0; x < container.width; x++) {
                let i = y * container.width + x
                if (x == 0) container.cells[i].classList.add('left-edge')
                if (x == 9) container.cells[i].classList.add('right-edge')
                if (y == container.height - 1) container.cells[i].classList.add('bottom-edge')
            }
        }
    }

    container.drawShape = (shape, pos, c, undraw = false) => {
        /** shape is stored on `size` x `size` grid */
        let size = Math.sqrt(shape.length)
        for (let y = 0; y < size; y++) {
            for (let x = 0; x < size; x++) {
                if (shape[y * size + x]) {
                    container.setCell(vec2(pos.x + x, pos.y + y), c, undraw)
                }
            }
        }
    }

    container.getShapeCells = (shape, pos) => {
        /** shape is stored on `size` x `size` grid */
        let size = Math.sqrt(shape.length)
        let cells = []
        for (let y = 0; y < size; y++) {
            for (let x = 0; x < size; x++) {
                if (shape[y * size + x]) {
                    cells.push(container.getCell(vec2(pos.x + x, pos.y + y)))
                }
            }
        }
        return cells
    }

    container.clear = () => {
        container.cells.forEach(c => c.className = '')
    }

    for (let i = 0; i < width * height; i++) {
        const div = document.createElement('div')
        container.appendChild(div)
    }

    container.cells = [...container.querySelectorAll('div')]

    container.markEdges()
}

document.addEventListener('DOMContentLoaded', () => {
    const board = document.querySelector('#grid')
    const preview = document.querySelector('#preview')

    const points = document.querySelector('#points')

    // on-screen buttons
    const start = document.querySelector('#start')
    const up = document.querySelector('#up')
    const down = document.querySelector('#down')
    const left = document.querySelector('#left')
    const right = document.querySelector('#right')

    createGrid(board)
    createGrid(preview, 4, 4)

    // tetrominoes and their rotations
    const r = [
        0, 0, 0,
        1, 1, 1,
        0, 0, 1,
    ]

    const l = [
        0, 0, 0,
        1, 1, 1,
        1, 0, 0,
    ]

    const s = [
        0, 1, 1,
        1, 1, 0,
        0, 0, 0,
    ]

    const z = [
        1, 1, 0,
        0, 1, 1,
        0, 0, 0,
    ]

    const t = [
        0, 0, 0,
        1, 1, 1,
        0, 1, 0,
    ]

    const i = [
        0, 0, 0, 0,
        1, 1, 1, 1,
        0, 0, 0, 0,
        0, 0, 0, 0,
    ]

    const o = [
        1, 1,
        1, 1,
    ]

    const tetros = [r, l, s, z, t, i, o]
    const offset = [1, 1, 0, 0, 1, 1, 0]

    let paused = true

    let current;
    let next;
    let rotation;
    let position;
    let tetro;

    spawn(random())

    function random() {
        return Math.floor(Math.random() * tetros.length)
    }

    function spawn(newCurrent) {
        current = newCurrent
        next = random()
        rotation = 0
        position = vec2(4, -1 * offset[current])
        tetro = tetros[current]
        draw()
        preview.clear()
        preview.drawShape(tetros[next], vec2(0, 0), 'tetro')
        if (board.getShapeCells(tetro, position).some(c => c !== undefined && c.classList.contains('frozen'))) {
            death()
        }
    }

    function draw(c = 'tetro') {
        board.drawShape(tetro, position, c)
    }

    function undraw(c = 'tetro') {
        board.drawShape(tetro, position, c, true)
    }

    let timer = setInterval(update, 1000);

    function update() {
        if (paused) return
        freeze()
        moveDown()
        score()
    }

    function touchingLeft() {
        return board.getShapeCells(tetro, position).some(c => c !== undefined && c.classList.contains('left-edge'))
    }

    function touchingRight() {
        return board.getShapeCells(tetro, position).some(c => c !== undefined && c.classList.contains('right-edge'))
    }

    function touchingBottom() {
        return board.getShapeCells(tetro, position).some(c => c !== undefined && c.classList.contains('bottom-edge'))
    }

    function moveDown() {
        if (touchingBottom() || board.getShapeCells( // if already touching bottom or
            tetro,                                   // any cell in the shape of this tetro
            vec2(position.x, position.y + 1)         // but on the next row
        ).some(c => c !== undefined &&               // some cells that fit on the board
                    c.classList.contains('frozen'))) // are colliding with a frozen cell
            return                                   // we do nothing

        undraw()
        position.y++
        draw()
    }

    function moveLeft() {
        if (touchingLeft() || board.getShapeCells(
            tetro,
            vec2(position.x - 1, position.y)
        ).some(c => c !== undefined &&
                    c.classList.contains('frozen'))) return

        undraw()
        position.x--
        draw()
    }

    function moveRight() {
        if (touchingRight() || board.getShapeCells(
            tetro,
            vec2(position.x + 1, position.y)
        ).some(c => c !== undefined &&
                    c.classList.contains('frozen'))) return

        undraw()
        position.x++
        draw()
    }

    /**
     * Math for different matrix rotations explained in matrix.txt.
     * I only need 90°: size*(size-1) - size*x + y
     */
    function rotate90(shape) {
        let rotated = []
        let size = Math.sqrt(shape.length)
        for (let y = 0; y < size; y++) {
            for (let x = 0; x < size; x++) {
                rotated.push(shape[size * (size - 1) - size * x + y])
            }
        }
        return rotated
    }

    function rotate() {
        undraw()
        let unrotated = tetro
        tetro = rotate90(tetro)

        if (board.getShapeCells(tetro, position).some(c => c !== undefined && c.classList.contains('frozen')) ||
            touchingLeft() && touchingRight()
        ) {
            tetro = unrotated
        }
        draw()
    }

    function freeze() {
        if (!touchingBottom() && !board.getShapeCells(
            tetro,
            vec2(position.x, position.y + 1)
        ).some(c => c !== undefined && c.classList.contains('frozen')))
            return

        draw('frozen')
        spawn(next)
        position.y-- // update() calls moveDown() right after freeze() so new spawns start 1 row too low
    }

    function death() {
        paused = true
        alert("Game Over\nFinal score: " + Number(points.innerHTML))
        board.clear()
        board.markEdges()
        start.innerHTML = 'Start'
        points.innerHTML = 0
        updateTimer()
    }

    function updateTimer() {
        clearInterval(timer)
        timer = setInterval(update, 1000 - Number(points.innerHTML))
    }

    function score() {
        let scored = false
        for (let y = 0; y < board.height; y++) {
            let rowFrozen = true
            for (let x = 0; x < board.width; x++) {
                if (!board.cells[y * board.width + x].classList.contains('frozen')) {
                    rowFrozen = false
                    break
                }
            }
            if (!rowFrozen) continue;

            scored = true
            points.innerHTML = Number(points.innerHTML) + 10
            undraw()
            const removed = board.cells.splice(y * board.width, board.width)
            removed.forEach(c => c.className = '')
            board.cells = removed.concat(board.cells)
            board.cells.forEach(c => board.appendChild(c))
            updateTimer()
            draw()
        }
        scored && board.markEdges()
    }

    function togglePause() {
        paused = !paused
        start.innerHTML = paused ? 'Resume' : 'Pause'
    }

    // buttons support

    start.addEventListener('click', togglePause)
    up.addEventListener('click', rotate)
    down.addEventListener('click', moveDown)
    left.addEventListener('click', moveLeft)
    right.addEventListener('click', moveRight)

    // keyboard support

    document.addEventListener('keydown', e => {
        if (e.code === 'Space') togglePause()
        if (e.code === 'Enter') togglePause()
        if (paused) return
        if (e.code === 'ArrowUp') rotate()
        if (e.code === 'ArrowDown') moveDown()
        if (e.code === 'ArrowLeft') moveLeft()
        if (e.code === 'ArrowRight') moveRight()
    })

    // touchscreen gestures support

    /** last `touchstart` event */
    let touch;

    document.addEventListener('touchstart', e => {
        let t = [...e.touches]
        if (t.length > 1) { // more than 1 finger
            togglePause()
            return
        }
        touch = t[0]
    })

    document.addEventListener('touchend', e => {
        if (paused) return;

        let deadzone = 40
        let touchend = [...e.changedTouches][0]

        if (touch.identifier !== touchend.identifier) return;

        if (touch.screenY > touchend.screenY + deadzone) rotate();
        if (touch.screenY < touchend.screenY - deadzone) moveDown();
        if (touch.screenX > touchend.screenX + deadzone) moveLeft();
        if (touch.screenX < touchend.screenX - deadzone) moveRight();

        if (touch.identifier === touchend.identifier) touch = null
    })
})
